from particles import variables
from multiprocessing import Process, Queue
import collections
from tqdm import tqdm
import sys
import numpy as np


class assignments(object):
    def __init__(self, ass_type, lhs, rhs):
        '''
        All kinds of assignments statements
        will be contained here
        '''

        self.ass_type = ass_type
        assert isinstance(lhs, list)
        assert isinstance(rhs, variables)
        for elems in lhs:
            assert isinstance(elems, variables), elems + \
                " should be a variable."
        self.nargs = len(lhs)
        self.lhs = lhs
        self.rhs = rhs

        assert self.rhs not in self.lhs, "RHS cannot be in LHS"

    def __str__(self):
        pass

    def __add__(self, x):
        return str(self) + x

    def __radd__(self, x):
        return x + str(self)


class mono_assign(assignments):
    def __init__(self, ass_type, lhs, rhs):
        assignments.__init__(self, ass_type, lhs, rhs)
        assert self.nargs == 1, "Cannot have more than one element : " + \
            rhs + " <= " + ass_type + ' ' + \
            ','.join(map(lambda x: str(x), lhs))
        self.ass_comms = {'not': 'NEGATE',
                          'alias': 'ALIAS'}

    def interpret(self):

        try:
            assert self.ass_type is not None
            self.command = self.ass_comms[self.ass_type]
        except KeyError:
            print("Maybe use one of  : " + ' | '.join(self.ass_comms))
            sys.exit()

    def __str__(self):
        self.interpret()
        return self.lhs[0] + ' ' + self.command + ' ' + self.rhs


class bi_assign(assignments):
    def __init__(self, ass_type, lhs, rhs):
        assignments.__init__(self, ass_type, lhs, rhs)
        assert self.nargs == 2, "Must have 2 elements: " + \
            rhs + " <= " + ass_type + ' ' + \
            ','.join(map(lambda x: str(x), lhs))
        self.ass_comms = {'xor': 'ADD',
                          'and': 'MULTIPLY',
                          'or': 'MAXIMUM',
                          'constand': 'MULTBYCONST'}

    def interpret(self):
        try:
            assert self.ass_type is not None
            self.command = self.ass_comms[self.ass_type]
        except KeyError:
            print("Maybe use one of  : " + ' | '.join(self.ass_comms))
            sys.exit()

    def __str__(self):
        self.interpret()
        lhs_args = self.lhs[0] + ' ' + self.lhs[1]
        return lhs_args + ' ' + self.command + ' ' + self.rhs


class tri_assign(assignments):
    def __init__(self, ass_type, lhs, rhs):
        assignments.__init__(self, ass_type, lhs, rhs)
        assert self.nargs == 3, "Must have 3 elements: " + \
            rhs + " <= " + ass_type + ' ' + \
            ','.join(map(lambda x: str(x), lhs))
        self.ass_comms = {'mux': 'SELECT'}

    def interpret(self):
        try:
            assert self.ass_type is not None
            self.command = self.ass_comms[self.ass_type]
        except KeyError:
            print("Maybe use one of  : " + ' | '.join(self.ass_comms))
            sys.exit()

    def __str__(self):
        self.interpret()
        lhs_args = self.lhs[0] + ' ' + self.lhs[1] + ' ' + self.lhs[2]
        return lhs_args + ' ' + self.command + ' ' + self.rhs


class mini_mod(object):
    def __init__(self, name, inputs, outputs, parallel=False):
        self.name = name
        self.inputs = inputs
        self.outputs = outputs
        self.commands = collections.OrderedDict()
        self.zero = variables(name='FALSE')
        self.one = variables(name='TRUE')
        self.parallel = parallel

    def add(self, op):
        try:
            if isinstance(op, mini_mod):
                self.commands[op.name] = op
            elif isinstance(op, assignments):
                self.commands[op.ass_type +
                              '_'.join(map(lambda x: str(x),
                                           op.lhs)) + '||' + op.rhs] = op
        except Exception:
            print("I am not quite sure, what you're trying\
                   to add. I will just exit!")

    def __str__(self):
        def local_str(output, idx, curr_module):
            output.put((idx, str(curr_module)))
            print "HI"
        vals = self.commands.values()

        if self.parallel:
            tgt_list = ['' for s in vals]
            output = Queue()
            proc_arr = []
            num_coms = len(vals)
            for i in tqdm(range(num_coms)):
                proc_arr.append(
                    Process(target=local_str, args=(output, i, vals[i])))
                proc_arr[i].start()
            for i in tqdm(range(num_coms)):
                proc_arr[i].join()
            results = [output.get() for p in proc_arr]
            for idx, val in enumerate(results):
                tgt_list[val[0]] = str(val[1])
        else:
            tgt_list = [str(x) for x in vals]
        return '\n'.join(tgt_list)


if __name__ == '__main__':
    a = variables('a')
    b = variables('b')
    c = variables('c')
    d = variables('d')
    t1 = mono_assign(ass_type='not', lhs=[a], rhs=b)
    print(t1)
    t2 = mono_assign(ass_type='alias', lhs=[b], rhs=d)
    print(t2)
    t3 = bi_assign(ass_type='or', lhs=[d, c], rhs=b)
    print(t3)

    print("MODULE below")
    t = mini_mod('addition', [a, c], b)
    t.add(t1)
    t.add(t2)
    t.add(t3)
    ts = mini_mod('addition', [a, c], b)
    print(t)



"""
Run low-level benchmark tests over specified combinations of 
gate, depth, input_type, context.
"""
import subprocess
import os
import re
from scripts.benchmark_utils import run_circuit, write_params_file, write_inputs_file

if not "SHEEP_HOME" in os.environ.keys():
    BASE_DIR = os.environ["HOME"]+"/SHEEP"
else:
    BASE_DIR = os.environ["SHEEP_HOME"]


CIRCUIT_FILE_DIR = BASE_DIR+"/benchmark_inputs/low_level/circuits"
if not os.path.exists(CIRCUIT_FILE_DIR):
    os.system("mkdir "+CIRCUIT_FILE_DIR)

INPUT_FILE_DIR = BASE_DIR+"/benchmark_inputs/low_level/inputs"
if not os.path.exists(INPUT_FILE_DIR):
    os.system("mkdir "+INPUT_FILE_DIR)    

DEBUG_FILE_DIR = BASE_DIR+"/benchmark_inputs/low_level/debug"
if not os.path.exists(DEBUG_FILE_DIR):
    os.system("mkdir "+DEBUG_FILE_DIR)    

def low_level_benchmarks(gates,types,contexts,max_depth=10):
    """
    call this function with lists of the Gates, input-types, and contexts to be benchmarked.
    this in turn calls run_single_benchmark for each combination.
    """
    for gate in gates:
        for input_type in types:
            bitwidth = get_bitwidth(input_type)
            signed = input_type.startswith("i")
            for depth in range(1,max_depth):
                circuit_file = CIRCUIT_FILE_DIR+"/circuit-"+gate+"-"+str(depth)+".sheep"

### inputs file depends on the gate now - most gates have depth+1 inputs, but SELECT and NEGATE have
### different requirements
                
                inputs_file = INPUT_FILE_DIR+"/inputs-"
                if gate == "SELECT":
                    inputs_file += "select-"
                elif gate == "NEGATE":
                    inputs_file += "1-to-1-"
                else:
                    inputs_file += "2-to-1-"
                inputs_file +=input_type+"-"+str(depth)+".inputs"
### loop over desired contexts
                for context in contexts:
                    print("Doing benchmark for %s %s %i %s" %
                          (context,gate,depth,input_type))
### run the test
                    results = run_circuit(circuit_file,
                                          inputs_file,
                                          input_type,
                                          context,
                                          "serial",
                                          params,
                                          "debug_"+context+".txt")


def params_for_level(context,level):
    """
    set parameters for a given context for a given level
    """
    if context == "HElib_Fp":
        param_dict = {"Levels": level+2}
        param_file = write_params_file(param_dict)
        return param_file
    elif context == "SEAL":
        param_dict = {
            1: {"N": 2048},
            2: {"N": 4096},
            3: {"N": 4096},
            4: {"N": 8192},
            5: {"N": 8192},
            6: {"N": 16384},
            7: {"N": 16384},
            8: {"N": 16384},
            9: {"N": 32768},                             
        }
        param_file = write_params_file(param_dict[level])
        return param_file
    else:
        return None
    
def scan_1(contexts=["HElib_Fp","SEAL"]):
    """
    one level of each gate, levels set to 1, 4, 8
    """
    for gate in ["ADD","SUBTRACT","MULTIPLY","SELECT","NEGATE"]:
        for input_type in ["bool","uint8_t","int8_t","uint16_t","int16_t","uint32_t","int32_t"]:
            for d in [1,4,8]:
                for context in contexts:
                    param_file = params_for_level(context,d)
                    circuit_file = CIRCUIT_FILE_DIR+"/circuit-"+gate+"-1.sheep"
### inputs file depends on the gate now - most gates have depth+1 inputs, but SELECT and NEGATE have
### different requirements                
                    inputs_file = INPUT_FILE_DIR+"/inputs-"
                    if gate == "SELECT":
                        inputs_file += "select-"
                    elif gate == "NEGATE":
                        inputs_file += "1-to-1-"
                    else:
                        inputs_file += "2-to-1-"
                    inputs_file +=input_type+"-1.inputs"
                    print("Doing benchmark for %s %s %s %i" %
                          (context,gate,input_type,d))
                    results = run_circuit(circuit_file,
                                          inputs_file,
                                          input_type,
                                          context,
                                          "serial",
                                          param_file,
                                          DEBUG_FILE_DIR+"/debug_"+gate+str(d)+input_type+".txt"
                                          )


def scan_2(contexts=["HElib_Fp","SEAL"]):
    """
    set levels set to 4, see where we go wrong
    """
    for gate in ["ADD","SUBTRACT","MULTIPLY","SELECT","NEGATE"]:
        for input_type in ["bool","uint8_t","int8_t","uint16_t","int16_t","uint32_t","int32_t"]:
            for d in range(1,9):
                for context in contexts:
                    param_file = params_for_level(context,4)
                    circuit_file = CIRCUIT_FILE_DIR+"/circuit-"+gate+"-"+str(d)+".sheep"
### inputs file depends on the gate now - most gates have depth+1 inputs, but SELECT and NEGATE have
### different requirements                
                    inputs_file = INPUT_FILE_DIR+"/inputs-"
                    if gate == "SELECT":
                        inputs_file += "select-"
                    elif gate == "NEGATE":
                        inputs_file += "1-to-1-"
                    else:
                        inputs_file += "2-to-1-"
                    inputs_file +=input_type+"-"+str(d)+".inputs"
                    print("Doing benchmark for %s %s %s %i" %
                          (context,gate,input_type,d))
                    results = run_circuit(circuit_file,
                                          inputs_file,
                                          input_type,
                                          context,
                                          "serial",
                                          param_file,
                                          DEBUG_FILE_DIR+"/debug_"+gate+str(d)+input_type+".txt"
                                          )    


def scan_3(contexts=["TFHE","HElib_F2","HElib_Fp","SEAL"]):
    """
    all tests, with the levels set appropriately
    """
    for gate in ["ADD","SUBTRACT","MULTIPLY","SELECT","NEGATE"]:
        for input_type in ["bool","uint8_t","uint16_t","uint32_t"]:
            for d in range(1,9):
                for context in contexts:
                    param_file = params_for_level(context,d)
                    circuit_file = CIRCUIT_FILE_DIR+"/circuit-"+gate+"-"+str(d)+".sheep"
### inputs file depends on the gate now - most gates have depth+1 inputs, but SELECT and NEGATE have
### different requirements                
                    inputs_file = INPUT_FILE_DIR+"/inputs-"
                    if gate == "SELECT":
                        inputs_file += "select-"
                    elif gate == "NEGATE":
                        inputs_file += "1-to-1-"
                    else:
                        inputs_file += "2-to-1-"
                    inputs_file +=input_type+"-"+str(d)+".inputs"
                    print("Doing benchmark for %s %s %s %i" %
                          (context,gate,input_type,d))
                    results = run_circuit(circuit_file,
                                          inputs_file,
                                          input_type,
                                          context,
                                          "serial",
                                          param_file,
                                          DEBUG_FILE_DIR+"/debug_"+gate+str(d)+input_type+".txt"
                                          )    



import sys
import uuid
import copy
import numpy as np
import utils

if sys.version_info[0] < 3:
    __VERSION__ = 2
else:
    __VERSION__ = 3


class variables(object):
    def __init__(self, name, nb=1, randomize_name=0):
        self.name = name
        if randomize_name > 0:
            self.name = self.name + uuid.uuid4().hex[:randomize_name].upper()

        self.nb = nb
        if __VERSION__ == 2:
            assert isinstance(name, basestring)
        else:
            assert isinstance(name, str)

    def __str__(self):
        return self.name

    def __add__(self, x):
        return str(self) + x

    def __radd__(self, x):
        return x + str(self)


class enc_vec(object):
    def __init__(self, name, nb=None, lst=None, randomize_name=4,
                 name_list=[]):
        self.randomize_name = randomize_name
        self.name = name
        self._lst = lst
        self.nb = nb
        self.size = nb
        assert (not (self.nb is None and self._lst is None)
                ), "Need to provide something"
        self.name_list = name_list
        self.create(name_list)
        self.name_list = self.get_variables()

    def create(self, name_list):
        '''
        An yet unbuilt vec constructor
        '''
        if len(name_list) == 0:
            self._lst = [variables(name=self.name + str(idx),
                                   randomize_name=self.randomize_name)
                         for idx in range(self.nb)]
        else:
            assert(len(name_list) == self.nb), "Name list \
            has differnt length " + \
                str(len(name_list)) + " , " + str(self.nb)
            self._lst = [variables(name=name_list[idx])
                         for idx in range(self.nb)]

    def get_input_dict(self, inp_var, write_file=False):
        '''
        Input : A list of valyues
        Output : A dictionary
        =========================
        Creates the input dictionary
        for each variable in the same order
        as _lst. This is necessary as some lists can
        have difficult names
        '''
        assert(isinstance(inp_var, list)), "Input should be a list"
        assert(len(inp_var) == self.nb), "length should be the same as vec"
        inp_dict = {'TRUE': 1, 'FALSE': 0}
        for idx in range(self.nb):
            inp_dict[self._lst[idx].name] = inp_var[idx]

        if write_file:
            return utils.write_inputs_file(inp_dict)
        else:
            return inp_dict

    def get_variables(self):
        return [self._lst[idx].name for idx in range(self.nb)]

    def __getitem__(self, item):
        '''
        Index an item
        '''
        target_lst = self._lst[item]
        if isinstance(item, slice):
            slice_length = len(target_lst)
            return enc_vec(name=self.name,
                           nb=slice_length,
                           lst=target_lst,
                           name_list=self.name_list[item])
        else:
            return target_lst

    def __setitem__(self, key, value):
        self._lst[key] = value
        self.name_list[key] = value

    def __add__(self, next_list):
        '''
        Concatenate two lists
        '''
        assert isinstance(next_list, enc_vec)
        new_name = self.name + '_' + next_list.name
        randomize_name = max(self.randomize_name, next_list.randomize_name)
        new_nb = self.nb + next_list.nb
        new_lst = self._lst + next_list._lst
        new_vec = enc_vec(name=new_name,
                          nb=new_nb,
                          lst=new_lst,
                          randomize_name=randomize_name,
                          name_list=self.name_list + next_list.name_list)
        return new_vec

    def __len__(self):
        return self.nb

    def append(self, obj):
        '''
        Append a variable at the end
        '''
        assert isinstance(
            obj, variables), "I'm not sure what you're trying to append here"
        self.nb += 1
        self._lst.append(obj)

    def push(self, obj):
        '''
        Append a variable at the beginning
        '''
        assert isinstance(
            obj, variables), "I'm not sure what you're trying to append here"
        self.nb += 1
        self.name_list = [obj.name] + self.name_list
        self._lst = [obj] + self._lst


class enc_mat(object):
    def __init__(self, name, size=None, randomize_name=4,
                 name_list=None):
        self.randomize_name = randomize_name
        self.name = name
        self.size = size
        self.nb = size[0] * size[1]
        assert (not (self.nb is None and self._lst is None)
                ), "Need to provide something"
        self._lst = []
        self.name_list = name_list
        self.create(name_list)

    def create(self, name_list):
        '''
        An yet unbuilt vec constructor
        '''

        if self.name_list is None:
            self.name_list = []
            for row_idx in range(self.size[0]):
                self._lst.append([variables(
                    name=self.name + str(row_idx) + str(col_idx),
                    randomize_name=self.randomize_name)
                    for col_idx in range(self.size[1])])
                self.name_list.append([var.name for var in self._lst[-1]])
        else:
            assert(len(self.name_list) == self.size[0])
            assert(len(self.name_list[0]) == self.size[1]), "Name list \
            has differnt length " + \
                str(len(self.name_list[0])) + " , " + str(self.nb)
            for row_idx in range(self.size[0]):
                self._lst.append([
                    variables(name=self.name_list[row_idx][col_idx])
                    for col_idx in range(self.size[1])])

    def transpose(self, order):
        def lst_trans(lst, order):
            np_arr = np.asarray(lst)
            np_trans = np_arr.transpose(order)
            return np_trans.shape, np_trans.tolist()
        new_mat = copy.deepcopy(self)
        new_mat.size, new_mat._lst = lst_trans(new_mat._lst, order)
        _, new_mat.name_list = lst_trans(new_mat.name_list, order)
        return new_mat

    def get_input_dict(self, inp_var, write_file=False):
        '''
        Input : A list of valyues
        Output : A dictionary
        =========================
        Creates the input dictionary
        for each variable in the same order
        as _lst. This is necessary as some lists can
        have difficult names
        '''
        assert(isinstance(inp_var, list)), "Input should be a list"
        assert(len(inp_var) == self.size[0] * self.size[1]
               ), str(len(inp_var)) + " ~ " + str(self.size[0] * self.size[1])
        inp_dict = {'TRUE': 1, 'FALSE': 0}
        var_list = self.get_variables()
        for idx, var in enumerate(var_list):
            inp_dict[var] = inp_var[idx]

        if write_file:
            return utils.write_inputs_file(inp_dict)
        else:
            return inp_dict

    def get_variables(self):
        var_list = []
        for row_idx in range(self.size[0]):
            var_list = var_list + \
                [self._lst[row_idx]
                    [col_idx].name for col_idx in range(self.size[1])]
        return var_list

    def __getitem__(self, item):
        '''
        Index an item
        '''
        assert not isinstance(item, slice), "NO advanced slicing"
        return enc_vec(name=self.name + '_sl_' + str(item),
                       name_list=self.name_list[item],
                       randomize_name=0,
                       nb=len(self.name_list[item]))

    def __setitem__(self, key, value):
        self._lst[key[0]][key[1]] = value
        self.name_list[key[0]][key[1]] = value.name

    def reorder(self, new_lst_, new_name_list_):
        '''
        It is not a safe method.
        It cannot be checked if they are still ordered
        '''
        self._lst = new_lst_
        self.name_list = new_name_list_

    def __add__(self, next_list):
        '''
        Concatenate two mats
        '''
        try:
            raise NotImplementedError
        except NotImplementedError:
            print("SORRY! Can't add enc mats")
            sys.exit()

    def __len__(self):
        return self.size[0]

    def append(self, obj):
        '''
        Append a variable at the end
        '''
        try:
            raise NotImplementedError
        except NotImplementedError:
            print("SORRY! Can't append to enc mats")
            sys.exit()

    def push(self, obj):
        '''
        Append a variable at the beginning
        '''
        try:
            raise NotImplementedError
        except NotImplementedError:
            print("SORRY! Can't push to enc mats")
            sys.exit()


class enc_tensor3(object):
    def __init__(self, name, size=None, randomize_name=4,
                 name_list=None):
        self.randomize_name = randomize_name
        self.name = name
        self.size = size
        self.nb = size[0] * size[1] * size[2]
        assert (not (self.nb is None and self._lst is None)
                ), "Need to provide something"
        self._lst = []
        self.name_list = []
        self.create(name_list)

    def create(self, name_list):
        '''
        An yet unbuilt vec constructor
        '''
        if name_list is None:
            for outer_idx in range(self.size[0]):
                temp_lst_ = []
                temp_name_list = []
                for row_idx in range(self.size[1]):
                    temp_lst_.append([variables(
                        name=self.name + str(outer_idx) +
                        str(row_idx) + str(col_idx),
                        randomize_name=self.randomize_name)
                        for col_idx in range(self.size[2])])
                    temp_name_list.append([var.name for var in temp_lst_[-1]])
                self._lst.append(temp_lst_)
                self.name_list.append(temp_name_list)
        else:
            assert(len(name_list) == self.size[0])
            assert(len(name_list[0]) == self.size[1])
            assert(len(name_list[2]) == self.size[2])
            for outer_idx in range(self.size[0]):
                for row_idx in range(self.size[1]):
                    self._lst.append([variables(
                        name=name_list[row_idx][col_idx])
                        for col_idx in range(self.size[2])])

    def transpose(self, order):
        def lst_trans(lst, order):
            np_arr = np.asarray(lst)
            np_trans = np_arr.transpose(order)
            return np_trans.shape, np_trans.tolist()
        new_mat = copy.deepcopy(self)
        new_mat.size, new_mat._lst = lst_trans(new_mat._lst, order)
        _, new_mat.name_list = lst_trans(new_mat.name_list, order)
        return new_mat

    def get_input_dict(self, inp_var, write_file=False):
        '''
        Input : A list of valyues
        Output : A dictionary
        =========================
        Creates the input dictionary
        for each variable in the same order
        as _lst. This is necessary as some lists can
        have difficult names
        '''
        assert(isinstance(inp_var, list)), "Input should be a list"
        assert(len(inp_var) == self.size[0] * self.size[1] * self.size[2]
               ), str(len(inp_var)) + " ~ " +\
            str(self.size[0] * self.size[1] * self.size[2])
        inp_dict = {'TRUE': 1, 'FALSE': 0}
        var_list = self.get_variables()
        for idx, var in enumerate(var_list):
            inp_dict[var] = inp_var[idx]

        if write_file:
            return utils.write_inputs_file(inp_dict)
        else:
            return inp_dict

    def get_variables(self):
        var_list = []
        for outer_idx in range(self.size[0]):
            for row_idx in range(self.size[1]):
                var_list = var_list +\
                    [self._lst[outer_idx][row_idx]
                        [col_idx].name for col_idx in range(self.size[2])]
        return var_list

    def __getitem__(self, item):
        '''
        Index an item
        '''
        assert not isinstance(item, slice), "NO advanced slicing"
        return enc_mat(name=self.name + '_sl_' + str(item),
                       name_list=self.name_list[item], randomize_name=0,
                       size=(self.size[1], self.size[2]))

    def __setitem__(self, key, value):
        self._lst[key[0]][key[1]][key[2]] = value
        self.name_list[key[0]][key[1]][key[2]] = value.name

    def reorder(self, new_lst_, new_name_list_):
        '''
        It is not a safe method.
        It cannot be checked if they are still ordered
        '''
        self._lst = new_lst_
        self.name_list = new_name_list_

    def __add__(self, next_list):
        '''
        Concatenate two mats
        '''
        try:
            raise NotImplementedError
        except NotImplementedError:
            print("SORRY! Can't add enc mats")
            sys.exit()

    def __len__(self):
        return self.size[0]

    def append(self, obj):
        '''
        Append a variable at the end
        '''
        try:
            raise NotImplementedError
        except NotImplementedError:
            print("SORRY! Can't append to enc mats")
            sys.exit()

    def push(self, obj):
        '''
        Append a variable at the beginning
        '''
        try:
            raise NotImplementedError
        except NotImplementedError:
            print("SORRY! Can't push to enc mats")
            sys.exit()



import sys
import networkx as nx
from networkx.drawing.nx_agraph import graphviz_layout, to_agraph
from PIL import Image


class node(object):

    def __init__(self, val, node_type, op='GOD'):
        self.val = val
        self.node_type = node_type
        self.mono = False
        self.bi = False
        if self.val in ['ALIAS', 'NEGATE']:
            self.mono = True
        else:
            self.bi = True
        self.children = []
        self.op = op
        self.name = self.val + "(" + self.op + ")"

    def __str__(self, level=0):
        ret = "\t" * level + repr(self.val) + "(" + self.op + ")\n"
        for child in self.children:
            ret += child.__str__(level + 1)
        return ret

    def __repr__(self):
        return '<tree node representation>'


class networkx_graph(object):
    def __init__(self, node, filename='test'):
        self.output = node
        self.filename = filename
        self.G = nx.DiGraph()
        self.labeldict = {}
        self.edge_label = {}
        self.short_names = {'NEGATE': '~',
                            'ALIAS': '=',
                            'MULTIPLY': '&',
                            'ADD': 'XR',
                            'MAXIMUM': '|',
                            'GOD': 'OUT',
                            'SELECT': 'SEL',
                            'MULTBYCONST': '&c'}
        self.G.add_node(self.construct(node))

    def construct(self, curr_node):
        if curr_node.op != 'INPUT' and curr_node.op != 'OUTPUT':
            self.G.add_node(
                curr_node.val, label=self.short_names[curr_node.op])
            self.labeldict[curr_node.val] = self.short_names[curr_node.op]
        elif curr_node.op == 'OUTPUT':
            self.G.add_node(
                curr_node.val,
                color='red',
                label=curr_node.val[:-3])
            self.labeldict[curr_node.val] = curr_node.val[:-3]
        else:
            self.G.add_node(
                curr_node.val,
                color='red',
                label=curr_node.val[:-3])
            self.labeldict[curr_node.val] = curr_node.val[:-3]
        for idx, child in enumerate(curr_node.children):
            target_node = self.construct(child)
            self.G.add_edge(target_node, curr_node.val,
                            label=str(idx))
            self.edge_label[(target_node, curr_node.val)
                            ] = target_node[-4:]

        return curr_node.val

    def draw(self):
        options = {
            'node_color': 'blue',
            'node_size': 10,
            'width': 1,
            'font_size': 9
        }
        nx.draw_networkx(self.G, labels=self.labeldict,
                         pos=graphviz_layout(self.G, prog='neato'),
                         with_label=True, **options)
        nx.draw_networkx_edges(self.G, pos=graphviz_layout(
            self.G, prog='neato'), arrowstyle='->', arrow=True)
        nx.draw_networkx_edge_labels(self.G, pos=graphviz_layout(
            self.G, prog='neato'), edge_labels=self.edge_label,
            arrowstyle='-')

        self.G.graph['graph'] = {'rankdir': 'TD', 'size': "20.,20"}
        self.G.graph['edges'] = {'arrowsize': '10.0'}
        G_agraph = to_agraph(self.G)
        G_agraph.layout(prog='dot')
        G_agraph.draw(self.filename.split('.')[0] + '.png')
        Image.open(self.filename.split('.')[0] + '.png').show()


def get_circuit_graph(filename):
    inps = []
    outs = []
    complete = node(val='OUTPUT',
                    node_type='COMPLETE')
    nodes = {'OUTPUT': complete}
    with open(filename, 'rb') as f:
        for idx, line in enumerate(f.readlines()):
            args = line.split()
            if idx <= 1:
                inps = args[1:]
                for vals in inps:
                    nodes[vals] = node(val=vals,
                                       node_type='INPUT',
                                       op='INPUT')
            elif idx == 2:
                outs = args[1:]
                for vals in outs:
                    nodes[vals] = node(val=vals,
                                       node_type='OUTPUT',
                                       op='OUTPUT')
                    nodes['OUTPUT'].children.append(nodes[vals])
            else:
                if len(args) == 5:
                    if args[-1] not in nodes.keys():
                        nodes[args[-1]] = node(val=args[-1],
                                               node_type='TRI_OP_OUT')
                    assert(args[0] in nodes.keys()), "Can't find " + args[0]
                    assert(args[1] in nodes.keys()), "Can't find " + args[1]
                    assert(args[2] in nodes.keys()), "Can't find " + args[2]
                    nodes[args[-1]].op = args[3]
                    nodes[args[-1]].children.append(nodes[args[0]])
                    nodes[args[-1]].children.append(nodes[args[1]])
                    nodes[args[-1]].children.append(nodes[args[2]])
                if len(args) == 4:
                    if args[-1] not in nodes.keys():
                        nodes[args[-1]] = node(val=args[-1],
                                               node_type='BI_OP_OUT')
                    assert(args[0] in nodes.keys()), "Can't find " + args[0]
                    assert(args[1] in nodes.keys()), "Can't find " + args[1]
                    nodes[args[-1]].op = args[2]
                    nodes[args[-1]].children.append(nodes[args[0]])
                    nodes[args[-1]].children.append(nodes[args[1]])
                elif len(args) == 3:
                    if args[-1] not in nodes.keys():
                        nodes[args[-1]] = node(val=args[-1],
                                               node_type='MONO_OP_OUT')
                    assert(args[0] in nodes.keys()), "Can't find " + args[0]
                    nodes[args[-1]].op = args[1]
                    nodes[args[-1]].children.append(nodes[args[0]])
    return nodes['OUTPUT']


if __name__ == '__main__':
    assert(len(sys.argv) == 2)
    complete_node = get_circuit_graph(sys.argv[-1])
    # print(complete_node)
    ng = networkx_graph(complete_node)
    ng.draw()



from particles import variables
from particles import enc_mat, enc_vec, enc_tensor3
import interactions as assignemnts
import functions
import circuit
import nn_layer
import reusable_modules
import interactions

__all__ = [variables, enc_mat, enc_vec, enc_tensor3,
           assignemnts, functions, nn_layer, reusable_modules,
           interactions, circuit]

import utils  # noqa
import create_graph  # noqa
